#include <allegro.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <stdio.h> /* For (v)snprintf */

#ifdef USE_ALFONT
#include <alfont.h>
extern ALFONT_FONT *alfont;
#endif

#include "config.h"
#include "chat.h"
#include "network.h"
#include "command.h"
#include "commandqueue.h"
#include "user.h"
#include "assert.h"
#include "rng.h"

extern Connection connection;
extern CommandQueue pending;
extern int can_connect, connected;
extern uint8_t userid;

#define MAX_MESSAGE_LENGTH 600
#define MAX_MESSAGE 150

#define SERVER_COLOUR  252
#define USER_COLOUR    253
#define TEXT_COLOUR    254
#define MESSAGE_COLOUR 255

typedef struct
 {
  char message[MAX_MESSAGE_LENGTH];
  int len;
  time_t birth;
  int show;
 } Message;

static Message message_start[MAX_MESSAGE];
static const Message *message_end = message_start+MAX_MESSAGE;
static Message *message_first = message_start, *message_last = message_start;
static int redraw = 1;

static const char *yes_no_msg[] = {"Signs point to yes.", "Yes.", "Reply hazy, try again.", "Without a doubt.",
 "My sources say no.", "As I see it, yes.", "You may rely on it.", "Concentrate and ask again.",
 "Outlook not so good.", "It is decidedly so.", "Better not tell you now.", "Very doubtful.", "Yes - definitely.",
 "It is certain.", "Cannot predict now.", "Most likely.", "Ask again later.", "My reply is no.", "Outlook good.",
 "Don't count on it.", "Yes, in due time.", "My sources say no.", "Definitely not.", "You will have to wait.",
 "I have my doubts.", "Outlook so so.", "Looks good to me!", "Who knows?", "Probably.", "Are you kidding?",
 "Go for it!", "Don't bet on it.", "Forget about it."};

static const char *size_msg[] = {"Fairly small.", "Larger than your puny mind can comprehend.", "Smaller than average",
 "Big enough, any larger would complicate matters.", "Of average size.", "Too small to be of any real use.",
 "It's not the size that counts, it's how you use it.", "It's a matter of personal preference."};

static const char *colour_msg[] = {"Red, with blue spots.", "All the colours of the rainbow, and then some.",
 "Several super-intelligent shades of the color blue.", "Plain boring white.", "Black as night.",
 "A nice earthy shade of brown.", "A healthy shade of green.", "Depends on the lighting conditions.",
 "I think I may be colour blind.", "Yellow, with green stripes.", "Pink. No wait! Blue!"};

static const char *where_msg[] = {"Behind you!", "Under your bed.", "Somewhere beneath Manhattan.", "Beyond the stars.",
 "Deep under the ocean.", "It was lost when Atlantis sank.", "Check behind your ears.",
 "It's probably under the sofa cushion with your lost change.", "Your neighbor 'borrowed' it.",
 "Did you check the lost and found box?", "It's somewhere around here.", "Sorry, it's lost forever.",
 "I don't think you'll ever find it."};

static const char *when_msg[] = {"Not in this lifetime.", "Tomorrow, maybe later.", "Very soon", "Probably never.",
 "Within the next several minutes.", "Possibly next week.", "It's going to be a few more decades, sorry.",
 "You missed it, it already happened!", "Five minutes ago.", "Yesterday.", "Right this very instant!"};

static const char *number_msg[] = {"The answer it most definitely 42.", "One. No more, no less.", "Sorry, they're all out.",
 "Enough for a small meal.", "Two or three. Possibly Four. Five at the most.", "Several dozen.", "Seventeen billion at least.",
 "More than you could possibly imagine.", "Not too many.", "Not nearly enough.", "As many as you want.", "A couple."};

void chatCommand(char *command)
 {
  const char *delim = " ";
  
  char *token = strtok(command, delim);
  
  if(!token)
   {
    chatAddMessage(0, "Expected a command name. Type \xff\xff/help\xff\xfc for help.");
    return;
   }
  
  if(!strcasecmp_custom(token, "8ball") || !strcasecmp_custom(token, "8"))
   {
    enum {YesNo, Colour, Size, Where, When, Number} type = YesNo;
    int question = 0;
    
    while((token = strtok(0, delim)))
     {
      question = 1;
      if(!strcasecmp_custom(token, "big") || !strcasecmp_custom(token, "small") || !strcasecmp_custom(token, "size"))
       {
        type = Size;
        break;
       }
      
      if(!strcasecmp_custom(token, "colour") || !strcasecmp_custom(token, "color") || !strcasecmp_custom(token, "hue"))
       {
        type = Colour;
        break;
       }
      
      if(!strcasecmp_custom(token, "when"))
       {
        type = When;
        break;
       }
      
      if(!strcasecmp_custom(token, "where"))
       {
        type = Where;
        break;
       }
      
      if(!strcasecmp_custom(token, "number") || !strcasecmp_custom(token, "many"))
       {
        type = Number;
        break;
       }
     }
    
    if(!question)
     {
      chatAddMessage(0, "How can I answer if you don't know the question?");
      return;
     }
    
    switch(type)
     {
      default:
      case YesNo:
       chatAddMessage(0, yes_no_msg[RNG()%(sizeof yes_no_msg / sizeof *yes_no_msg)]);
      break;
      
      case Size:
       chatAddMessage(0, size_msg[RNG()%(sizeof size_msg / sizeof *size_msg)]);
      break;
      
      case Colour:
       chatAddMessage(0, colour_msg[RNG()%(sizeof colour_msg / sizeof *colour_msg)]);
      break;
      
      case Where:
       chatAddMessage(0, where_msg[RNG()%(sizeof where_msg / sizeof *where_msg)]);
      break;
      
      case When:
       chatAddMessage(0, when_msg[RNG()%(sizeof when_msg / sizeof *when_msg)]);
      break;
      
      case Number:
       chatAddMessage(0, number_msg[RNG()%(sizeof number_msg / sizeof *number_msg)]);
      break;
     }
    
    return;
   }
  
  if(!strcasecmp_custom(token, "restart") || !strcasecmp_custom(token, "reset") || !strcasecmp_custom(token, "r"))
   {
    if(strtok(0, delim))
     {
      chatAddMessage(0, "Unexpected argument. Type \xff\xff/help restart\xff\xfc for help.");
     }
    else
     {
      if(can_connect && connected)
       {
        Command command;
        commandInit(&command, Restart);
        
        if(!commandPutU8(&command, 1) || !commandNetworkSend(&command, &connection))
         chatAddMessage(0, "Error sending restart command.");
        
        commandUninit(&command);
       }
      else chatAddMessage(0, "You're not connected to a server.");
     }
    
    return;
   }
  
  if(!strcasecmp_custom(token, "motd"))
   {
    if(!(token = strtok(0, delim)))
     {
      chatAddMessage(0, "Expected password. Type \xff\xff/help motd\xff\xfc for help.");
     }
    else
     {
      const char *message = strtok(0, "");
      
      if(can_connect && connected)
       {
        Command command;
        commandInit(&command, Motd);
        
        if(!commandPutString(&command, token) || !commandPutString(&command, message?message:"") || !commandNetworkSend(&command, &connection))
         chatAddMessage(0, "Error sending motd command.");
        
        commandUninit(&command);
       }
      else chatAddMessage(0, "You're not connected to a server.");
     }
    
    return;
   }
  
  if(!strcasecmp_custom(token, "kick") || !strcasecmp_custom(token, "k"))
   {
    if((token = strtok(0, delim)))
     {
      char *end = token;
      int id = strtol(token, &end, 10);
      
      if(strtok(0, delim))
       {
        chatAddMessage(0, "Unexpected argument. Type \xff\xff/help kick\xff\xfc for help.");
        return;
       }
      
      if(end != token)
       {
        if(end != token+strlen(token))
         {
          chatAddMessage(0, "Unable to parse ID number.");
          return;
         }
        
        if(!userValid(id))
         {
          chatAddMessage(0, "Invalid ID number.");
          return;
         }
       }
      else if(!(id = userFind(token)))
       {
        chatAddMessage(0, "Invalid user.");
        return;
       }
      
      if(can_connect && connected)
       {
        Command command;
        commandInit(&command, Kick);
        
        if(!commandPutU8(&command, id) || !commandNetworkSend(&command, &connection))
         chatAddMessage(0, "Error sending kick command.");
        
        commandUninit(&command);
       }
      else chatAddMessage(0, "You're not connected to a server.");
     }
    else
     {
      chatAddMessage(0, "Expected user name or id number. Type \xff\xff/help kick\xff\xfc for help.");
     }
    
    return;
   }
  
  if(!strcasecmp_custom(token, "user") || !strcasecmp_custom(token, "users") || !strcasecmp_custom(token, "list") || !strcasecmp_custom(token, "l") || !strcasecmp_custom(token, "u"))
   {
    if((token = strtok(0, delim)))
     {
      chatAddMessage(0, "Unexpected argument. Type \xff\xff/help user\xff\xfc for help.");
     }
    else
     {
      int c;
      chatAddMessage(0, "Users:");
      for(c = 1; c < 255; ++c) if(userValid(c))
       {
        char buf[128];
        snprintf(buf, sizeof buf, " %3d - %s", c, userGetName(c));
        chatAddMessage(0, buf);
       }
     }
    
    return;
   }
  
  if(!strcasecmp_custom(token, "help") || !strcasecmp_custom(token, "h") || !strcasecmp_custom(token, "?"))
   {
    if((token = strtok(0, delim)))
     {
      if(strtok(0, delim))
       {
        chatAddMessage(0, "Unexpected argument. Type \xff\xff/help help\xff\xfc for help.");
        return;
       }
       
      if(!strcasecmp_custom(token, "kick") || !strcasecmp_custom(token, "k"))
       {
        chatAddMessage(0, "Kick:");
        chatAddMessage(0, "Usage: \xff\xff/kick [user-name|user-id]");
        chatAddMessage(0, "Aliases: kick, k");
        chatAddMessage(0, "Description: Votes for a user to leave. User will be kicked if over 50% of users vote for that user to be kicked. You can only vote for one user at a time, voting a second time will replace your previous vote. You can type part of a name as long as only one user matches.");
        return;
       }
      
      if(!strcasecmp_custom(token, "restart") || !strcasecmp_custom(token, "reset") || !strcasecmp_custom(token, "r"))
       {
        chatAddMessage(0, "Restart:");
        chatAddMessage(0, "Usage: \xff\xff/restart");
        chatAddMessage(0, "Aliases: restart, reset, r");
        chatAddMessage(0, "Description: Votes for the server to restart. Will restart if over 50% of users vote for a restart.");
        return;
       }
      
      if(!strcasecmp_custom(token, "help") || !strcasecmp_custom(token, "h") || !strcasecmp_custom(token, "?"))
       {
        chatAddMessage(0, "Help:");
        chatAddMessage(0, "Usage: \xff\xff/help [command]");
        chatAddMessage(0, "Aliases: help, h, ?");
        chatAddMessage(0, "Description: If you're reading this, you know how to use the help command.");
        return;
       }
      
      if(!strcasecmp_custom(token, "user") || !strcasecmp_custom(token, "users") || !strcasecmp_custom(token, "list") || !strcasecmp_custom(token, "u") || !strcasecmp_custom(token, "l"))
       {
        chatAddMessage(0, "User:");
        chatAddMessage(0, "Usage: \xff\xff/user");
        chatAddMessage(0, "Aliases: user, users, list, u, l");
        chatAddMessage(0, "Description: Lists all of the users, with their id numbers.");
        return;
       }
      
      if(!strcasecmp_custom(token, "8ball") || !strcasecmp_custom(token, "8"))
       {
        chatAddMessage(0, "8Ball:");
        chatAddMessage(0, "Usage: \xff\xff/8ball [question]");
        chatAddMessage(0, "Aliases: 8ball, 8");
        chatAddMessage(0, "Description: The magic 8 Ball knows all. Ask, and it will tell.");
        return;
       }
      
      if(!strcasecmp_custom(token, "motd"))
       {
        chatAddMessage(0, "MOTD:");
        chatAddMessage(0, "Usage: \xff\xff/motd [password] [new message]");
        chatAddMessage(0, "Aliases: motd");
        chatAddMessage(0, "Description: Sets the 'Message of the Day'. Will be seen by all joining users.");
        return;
       }
      
      chatAddMessage(0, "Unknown command. Type \xff\xff/help\xff\xfc to list availible commands.");
     }
    else
     {
      chatAddMessage(0, "Type \xff\xff/help [commandname]\xff\xfc for additional help. Valid commands are:");
      chatAddMessage(0, " motd restart user kick help 8ball");
     }
    
    return;
   }
  
  chatAddMessage(0, "Unknown command. Type \xff\xff/help\xff\xfc to list availible commands.");
 }

void chatAddMessage(const char *name, const char *message)
 {
  if(message && *message)
   {
    if(name)
     snprintf(message_last->message, MAX_MESSAGE_LENGTH, "\xff%c%s\xff%c: \xff%c%s", USER_COLOUR, *name?name:"???", TEXT_COLOUR, MESSAGE_COLOUR, message);
    else snprintf(message_last->message, MAX_MESSAGE_LENGTH, "\xff%c%s", SERVER_COLOUR, message);
    
    message_last->len = strlen(message_last->message);
    message_last->birth = time(0);
    message_last->show = 2;
    
    if(++message_last == message_end)
     message_last = message_start;
    
    if(message_last == message_first)
     {
      if(++message_first == message_end)
       message_first = message_start;
     }
    
    redraw = 1;
   }
 }

int charLength(unsigned char c)
 {
  char buf[2];
  buf[0] = c;
  buf[1] = 0;
  
  #ifdef USE_ALFONT
  if(alfont)
   return alfont_text_length(alfont, buf);
  else
  #endif
   return text_length(font, buf);
 }

int charHeight(void)
 {
  #ifdef USE_ALFONT
  if(alfont)
   return alfont_text_height(alfont);
  else
  #endif
   return text_height(font);
 }

int wordLength(const char *text, int width)
 {
  int len = 0;
  for(; *text && *text != ' '; ++text)
   if(*text == (char)255 && text[1])
    ++text;
   else
    {
     int l = charLength(*text);
     
     if(l + len > width)
      {
       if(len)
        return len;
       
       return width;
      }
     
     len += l;
    }
  
  return len;
 }

const char* wordEnd(const char *text, int x, int width)
 {
  for(; *text && *text != ' '; ++text)
   if(*text == (char)255 && text[1])
    ++text;
   else
    {
     int l = charLength(*text);
     
     if(x + l > width)
      return text;
     
     x += l;
    }
  
  return text;
 }


int textHeight(const char *text, int width)
 {
  int x = width, y = 2;
  
  while(*text)
   {
    int length = wordLength(text, width);
    
    if(x + length >= width)
     {
      x = 0;
      y += charHeight()+2;
      length = wordLength(text, width);
     }
    
    text = wordEnd(text, x, width);
    x += length;
    
    for(; *text == ' '; ++text)
     x += charLength(' ');
   }
  
  return y+2;
 }

const char *wordDraw(BITMAP *bmp, const char *text, int x, int y, int width, unsigned char *colour)
 {
  int bg = makecol32(0, 0, 0);
  char buf[2];
  buf[1] = 0;
  
  assert(bmp && text && colour);
  
  for(;*text && *text != ' '; ++text)
   if(*text == (char)255 && text[1])
    *colour = *++text;
   else
    {
     int c, len;
     buf[0] = *text;
     len = charLength(*text);
     
     switch(*colour)
      {
       case SERVER_COLOUR:
        c = makecol32(255, 255, 0);
       break;
       
       case USER_COLOUR:
        c = makecol32(0, 255, 255);
       break;
       
       case MESSAGE_COLOUR:
        c = makecol32(192, 192, 192);
       break;
       
       case TEXT_COLOUR:
       default:
        c = makecol32(255, 255, 255);
       break;
      }
     
     if(x+len > width)
      return text;
     
     #ifdef USE_ALFONT
     if(alfont)
      {
       alfont_textout_aa_ex(bmp, alfont, buf, x+2, y+2, bg, bg);
       alfont_textout_aa_ex(bmp, alfont, buf, x+1, y+1, c, bg);
      }
     else
     #endif
      {
       #ifdef OLD_ALLEGRO
       textout(bmp, font, buf, x+2, y+1, bg);
       textout(bmp, font, buf, x+1, y+2, bg);
       textout(bmp, font, buf, x+1, y+1, c);
       #else
       textout_ex(bmp, font, buf, x+2, y+1, bg, -1);
       textout_ex(bmp, font, buf, x+1, y+2, bg, -1);
       textout_ex(bmp, font, buf, x+1, y+1, c, -1);
       #endif
      } 
     
     x += len;
    }
  
  return text;
 }

int textDraw(BITMAP *bmp, const char *text, int width, int y)
 {
  int x = width;
  unsigned char c = TEXT_COLOUR;
  assert(bmp && text);
  
  y -= charHeight();
  
  while(*text)
   {
    int length = wordLength(text, width-x);
    
    if(x + length >= width)
     {
      x = 0;
      
      if((y += charHeight()+2) >= bmp->h)
       return y;
      
      length = wordLength(text, width);
     }
    
    text = wordDraw(bmp, text, x, y, width, &c);
    x += length;
    
    for(; *text == ' '; ++text)
     x += charLength(' '); /* length of a space. */
   }
  
  return y+charHeight()+4;
 }

BITMAP * chatOptions(BITMAP *bmp, int btn, int key, int x, int y)
 {
  static char buf[512] = {0};
  static unsigned int buf_pos = 0;
  
  if(!bmp)
   {
    if(!(bmp = create_bitmap_ex(32, screen_w, screen_h-42)))
     return 0;
    
    redraw = 1;
   }
  
  if(key)
   {
    if((key >> 8) == KEY_BACKSPACE)
     {
      if(buf_pos)
       {
        --buf_pos;
        buf[buf_pos] = 0;
        redraw = 1;
       }
     }
    else if((key >> 8) == KEY_ENTER || (key >> 8) == KEY_ENTER_PAD)
     {
      if(buf_pos)
       {
        if(*buf == '/')
         chatCommand(buf+1);
        else if(can_connect && connected)
         {
          Command command;
          commandInit(&command, Chat);
          
          
          if(!commandPutU8(&command, userid) || !commandPutString(&command, buf) || !commandNetworkSend(&command, &connection))
           chatAddMessage(0, "Error sending chat message.");
          
          if(!commandQueuePushBack(&pending, &command))
           { /* We can't store it. We'll display it now, but it might be out of
                order, and the client will complain about the server accepting
                a command we didn't send when it accepts our message. */
            chatAddMessage(userGetName(userid), buf);
            commandUninit(&command);
           }
         }
        else
         chatAddMessage(0, "You're not connected to a server.");
        
        buf_pos = 0;
        buf[0] = 0;
        redraw = 1;
       }
     }
    else if(isprint(key&0xff))
     {
      if(buf_pos < sizeof buf / sizeof *buf - 1)
       {
        buf[buf_pos] = key&0xff;
        ++buf_pos;
        buf[buf_pos] = 0;
        redraw = 1;
       }
     }
   }
  
  if(redraw && !key)
   {
    int y = bmp->h - textHeight(buf, bmp->w);
    const Message *m = message_last;
    clear_to_color(bmp, makecol32(255, 0, 255));
    
    textDraw(bmp, buf, bmp->w, y);
    
    while(y > 0 && m != message_first)
     {
      if(--m < message_start)
       m = message_end - 1;
      
      textDraw(bmp, m->message, bmp->w, y -= textHeight(m->message, bmp->w));
     }
    
    redraw = 0;
    return bmp;
   }
  
  return 0;
 }

extern void redrawCanvas(BITMAP *);

int chatRedrawOverlay(BITMAP *buffer, BITMAP *canvas, int redraw)
 {
  int y = 0;
  time_t current_time = time(0);
  Message *m;
  int shown = 0;
  
  for(m = message_first ;; ++m)
   {
    if(m == message_end)
     m = message_start;
    
    if(m == message_last)
     break;
    
    if(m->show == 2)
     {
      m->show = 1;
      redraw = 1;
      shown = 1;
     }
    else if(!shown && m->show == 1 && current_time > (int)(m->birth + m->len*0.05 + 0.5))
     {
      m->show = 0;
      redraw = 1;
     }
    else if(m->show == 1)
     {
      if(shown)
       {
        m->birth = current_time;
       }
      else shown = 1;
     }
   }
  
  if(redraw)
   {
    redrawCanvas(canvas);
    
    for(m = message_first ;; ++m)
     {
      if(m == message_end)
       m = message_start;
      
      if(m == message_last)
       break;
      
      if(m->show) /* No need to redraw the region; the whole screen will be redrawn. */
       y = textDraw(buffer, m->message, screen_w, y);
     }
   }
  
  return redraw;
 }
