#ifdef BERKSOCK
/* Winsock and Berkeley sockets are almost the same,
   so you can pretty much getaway with just a few defines. */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <unistd.h>
#include <netdb.h>
#include <arpa/inet.h>
typedef int SOCKET;
#define INVALID_SOCKET -1
#define SOCKET_ERROR -1
#define closesocket(x) close(x)
#else
#include <winsock.h>
typedef int socklen_t;
#endif

#include <stdlib.h>
#include <string.h> /* Needed for memset() */
#include <time.h>

#include "mem.h"
#include "assert.h"
#include "network.h"

static Connection *start = 0, *end = 0, *alloc = 0;
static SOCKET server = INVALID_SOCKET;

int networkInit(void)
 {
  #ifdef BERKSOCK
  return 1;
  #else
  WSADATA wsaData;
  return WSAStartup(MAKEWORD(1, 0), &wsaData) == 0;
  #endif
 }

int networkServerInit(uint16_t port)
 {
  struct sockaddr_in address;
  int address_size = sizeof address;
  const int option_on = 1;
  
  if(INVALID_SOCKET == (server = socket(AF_INET, SOCK_STREAM, 0)))
   return 0;
  
  address.sin_family = AF_INET;
  address.sin_port = htons(port);
  address.sin_addr.s_addr = htonl(INADDR_ANY);
  memset(&(address.sin_zero), 0, sizeof address.sin_zero);
  
  setsockopt(server, SOL_SOCKET, SO_REUSEADDR, (const char *)&option_on, sizeof option_on);
  
  if(SOCKET_ERROR == bind(server, (struct sockaddr *)&address, address_size)
     || SOCKET_ERROR == listen(server, 5) || !(start = end = (Connection *)malloc(sizeof(Connection) * 16)))
   {
    closesocket(server);
    return 0;
   }
  
  alloc = start + 16;
  return 1;
 }

void networkServerUninit(void)
 {
  if(server != INVALID_SOCKET)
   {
    closesocket(server);
    server = INVALID_SOCKET;
    
    assert(start && end && alloc);
    
    while(end != start)
     networkKillConnection(end-1);
    
    free(start);
    start = end = alloc = 0;
   }
 }

void networkUninit(void)
 {
  #ifndef BERKSOCK
  WSACleanup();
  #endif
 }

int networkSendBuffer(Connection *c, const uint8_t *start, const uint8_t *end)
 {
  assert(c && start <= end);
  
  while(c->send_alloc - c->send_end < end - start)
   {
    uint8_t *old = c->send_start;
    if(!(c->send_start = (uint8_t *)realloc(old, (c->send_alloc - old)*2)))
     {
      c->send_start = old;
      return 0;
     }
    
    c->send_end = c->send_end - old + c->send_start;
    c->send_alloc = (c->send_alloc - old)*2 + c->send_start;
   }
  
  memcpy(c->send_end, start, end - start);
  c->send_end += end - start;
  return 1;
 }

uint16_t networkWaitingBytes(Connection *c)
 {
  assert(c);
  return c->recv_end - c->recv_start;
 }

const uint8_t *networkGetReadBuffer(Connection *c)
 {
  return c->recv_start;
 }

void networkSkipWaitingBytes(Connection *c, uint16_t count)
 {
  if(count && c->recv_start != c->recv_end)
   {
    if(count >= c->recv_end - c->recv_start)
     c->recv_end = c->recv_start;
    else
     {
      memmove(c->recv_start, c->recv_start + count, c->recv_end - c->recv_start - count);
      c->recv_end -= count;
     }
   }
 }

int networkConnectToServer(Connection *c, const char *name, uint16_t port)
 {
  SOCKET sock;
  struct hostent *he;
  struct sockaddr_in addr;
  
  assert(c && name && port);
  
  c->complete = 1;
  
  if(!(he=gethostbyname(name)))
   return 0;
  
  if(INVALID_SOCKET == (sock = socket(AF_INET, SOCK_STREAM, 0)))
   return 0;
  
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr = *((struct in_addr *)he->h_addr);
  memset(&(addr.sin_zero), 0, 8);
  
  if(SOCKET_ERROR == connect(sock, (struct sockaddr *)&addr, sizeof(struct sockaddr)))
   {
    closesocket(sock);
    return 0;
   }
  
  if(!(c->send_start = c->send_end = (uint8_t *)malloc(256)))
   {
    closesocket(sock);
    return 0;
   }
  
  c->send_alloc = c->send_start + 256;
  
  if(!(c->recv_start = c->recv_end = (uint8_t *)malloc(256)))
   {
    free(c->send_start);
    closesocket(sock);
    return 0;
   }
  
  strncpy(c->address, inet_ntoa(addr.sin_addr), sizeof c->address);
  c->recv_alloc = c->recv_start + 256;
  
  c->sock = sock;
  c->complete = 0;
  c->birth = c->last_reception = time(0);
  c->sent = c->recieved = 0;
  c->custom_data = 0;
  
  return 1;
 }

void networkKillConnection(Connection *c)
 {
  assert(c);
  
  free(c->send_start);
  free(c->recv_start);
  
  if(c->custom_data)
   free(c->custom_data);
  
  closesocket(c->sock);
  
  if(c >= start && c < end)
   {
    /* Need to actually shift the connections, changing the order would mess up updating. */
    for(; c+1 < end; ++c)
     *c = c[1];
    
    --end;
   }
 }

static Connection * networkCreateConnection(SOCKET sock, struct sockaddr_in *addr, socklen_t *addre_size)
 {
  assert(sock != INVALID_SOCKET);
  
  if(end == alloc)
   {
    Connection *old = start;
    if(!(start = (Connection *)realloc(old, sizeof(Connection) * (alloc - start) * 2)))
     {
      start = old;
      return 0;
     }
    
    end = end - old + start;
    alloc = (alloc - old)*2 + start;
   }
  
  strncpy(end->address, inet_ntoa(addr->sin_addr), sizeof end->address);
  end->sock = sock;
  end->birth = end->last_reception = time(0);
  end->sent = end->recieved = 0;
  end->custom_data = 0;
  end->complete = 0;
  
  if(!(end->send_start = end->send_end = (uint8_t *)malloc(256)))
   return 0;
  
  end->send_alloc = end->send_start + 256;
  
  if(!(end->recv_start = end->recv_end = (uint8_t *)malloc(256)))
   {
    free(end->send_start);
    return 0;
   }
  
  end->recv_alloc = end->recv_start + 256;
  
  return end++;
 }

int socketWritable(SOCKET sock)
 {
  struct timeval timeout = {0, 0};
  fd_set fd;
  FD_ZERO(&fd);
  FD_SET(sock, &fd);
  #ifdef BERKSOCK
  return select(sock + 1, 0, &fd, 0, &timeout);
  #else
  return select(0, 0, &fd, 0, &timeout);
  #endif
 }

int socketReadable(SOCKET sock)
 {
  struct timeval timeout = {0, 0};
  fd_set fd;
  FD_ZERO(&fd);
  FD_SET(sock, &fd);
  #ifdef BERKSOCK
  return select(sock + 1, &fd, 0, 0, &timeout);
  #else
  return select(0, &fd, 0, 0, &timeout);
  #endif
 }

int networkUpdateConnection(Connection *p)
 {
  if(p->send_start != p->send_end && socketWritable(p->sock))
   {
    int len;
    
    if(0 >= (len = send(p->sock, (const char *)p->send_start, p->send_end - p->send_start > 512 ? 512 : p->send_end - p->send_start, 0)))
     {
      p->complete = 1;
      return 0;
     }
    else
     {
      memmove(p->send_start, p->send_start + len, p->send_end - p->send_start - len);
      p->send_end -= len;
      p->last_reception = time(0);
      p->sent += len;
     }
   }
  
  if(!p->complete && socketReadable(p->sock))
   {
    int len;
    
    while(p->recv_alloc - p->recv_end < 256)
     {
      uint8_t *old = p->recv_start;
      
      if(!(p->recv_start = (uint8_t *)realloc(old, sizeof *old * (p->recv_alloc - old) * 2)))
       {
        p->recv_start = old;
        p->complete = 1;
        return 1;
       }
      
      p->recv_end = p->recv_end - old + p->recv_start;
      p->recv_alloc = (p->recv_alloc - old)*2 + p->recv_start;
     }
    
    if(0 >= (len = recv(p->sock, (char *)p->recv_end, p->recv_alloc - p->recv_end, 0)))
     {
      p->complete = 1;
      return 0;
     }
    
    p->recv_end += len;
    p->last_reception = time(0);
    p->recieved += len;
   }
  
  return 1;
 }

void networkUpdateServer(int(*connect)(Connection *, void **), int(*update)(Connection *, void **), void(*disconnect)(Connection *, void **))
 {
  struct sockaddr_in address;
  socklen_t address_size = sizeof address;
  SOCKET sock;
  Connection *p;
  
  if(socketReadable(server) && INVALID_SOCKET != (sock = accept(server, (struct sockaddr *)&address, &address_size)))
   {
    if(!(p = networkCreateConnection(sock, &address, &address_size)))
     closesocket(sock);
    else if(connect && !(*connect)(p, &(p->custom_data)))
     networkKillConnection(p);
   }
  
  for(p = start; p != end; ++p) if((p->complete && p->send_start == p->send_end) || (time(0) > p->last_reception + (p->complete ? 10 : 300)))
   {
    if(disconnect && p->custom_data)
     (*disconnect)(p, &(p->custom_data));
    
    networkKillConnection(p--);
   }
  else
   {
    if(!networkUpdateConnection(p))
     p->complete = 1;
    else if(!p->complete && update && !(*update)(p, &(p->custom_data)))
     p->complete = 1;
   }
 }

float networkUploadSpeed(const Connection *c)
 {
  float age = time(0) - c->birth;
  if(age < 10) age = 10;
  return (float)c->sent/age;
 }

float networkDownloadSpeed(const Connection *c)
 {
  float age = time(0) - c->birth;
  if(age < 10) age = 10;
  return (float)c->recieved/age;
 }

uint16_t networkClients(void)
 {
  return end-start;
 }

Connection *networkClient(uint16_t i)
 {
  if(i < end-start)
   return start + i;
  else return 0;
 }

